/*
	des.c
	Copyright (C) 2000 David Fifield

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "des.h"

static unsigned char key_perm[56] = {
	56, 48, 40, 32, 24, 16,  8,  0, 57, 49, 41, 33, 25, 17,  9,  1,
	58, 50, 42, 34, 26, 18, 10,  2, 59, 51, 43, 35, 62, 54, 46, 38,
	30, 22, 14,  6, 61, 53, 45, 37, 29, 21, 13,  5, 60, 52, 44, 36,
	28, 20, 12,  4, 27, 19, 11,  3
};

static unsigned char comp_perm[48] = {
	13, 16, 10, 23,  0,  4,  2, 27, 14,  5, 20,  9, 22, 18, 11,  3,
	25,  7, 15,  6, 26, 19, 12,  1, 40, 51, 30, 36, 46, 54, 29, 39,
	50, 44, 32, 47, 43, 48, 38, 55, 33, 52, 45, 41, 49, 35, 28, 31
};

/*
static unsigned char p_box_perm[32] = {
	15,  6, 19, 20, 28, 11, 27, 16,  0, 14, 22, 25,  4, 17, 30,  9,
	 1,  7, 23, 13, 31, 26,  2,  8, 18, 12, 29,  5, 21, 10,  3, 24
};
*/

unsigned char s_box[4][64] = {
	{
		0xEF, 0x03, 0x41, 0xFD, 0xD8, 0x74, 0x1E, 0x47,
		0x26, 0xEF, 0xFB, 0x22, 0xB3, 0xD8, 0x84, 0x1E,
		0x39, 0xAC, 0xA7, 0x60, 0x62, 0xC1, 0xCD, 0xBA,
		0x5C, 0x96, 0x90, 0x59, 0x05, 0x3B, 0x7A, 0x85,
		0x40, 0xFD, 0x1E, 0xC8, 0xE7, 0x8A, 0x8B, 0x21,
		0xDA, 0x43, 0x64, 0x9F, 0x2D, 0x14, 0xB1, 0x72,
		0xF5, 0x5B, 0xC8, 0xB6, 0x9C, 0x37, 0x76, 0xEC,
		0x39, 0xA0, 0xA3, 0x05, 0x52, 0x6E, 0x0F, 0xD9
	}, {
		0xA7, 0xDD, 0x0D, 0x78, 0x9E, 0x0B, 0xE3, 0x95,
		0x60, 0x36, 0x36, 0x4F, 0xF9, 0x60, 0x5A, 0xA3,
		0x11, 0x24, 0xD2, 0x87, 0xC8, 0x52, 0x75, 0xEC,
		0xBB, 0xC1, 0x4C, 0xBA, 0x24, 0xFE, 0x8F, 0x19,
		0xDA, 0x13, 0x66, 0xAF, 0x49, 0xD0, 0x90, 0x06,
		0x8C, 0x6A, 0xFB, 0x91, 0x37, 0x8D, 0x0D, 0x78,
		0xBF, 0x49, 0x11, 0xF4, 0x23, 0xE5, 0xCE, 0x3B,
		0x55, 0xBC, 0xA2, 0x57, 0xE8, 0x22, 0x74, 0xCE
	}, {
		0x2C, 0xEA, 0xC1, 0xBF, 0x4A, 0x24, 0x1F, 0xC2,
		0x79, 0x47, 0xA2, 0x7C, 0xB6, 0xD9, 0x68, 0x15,
		0x80, 0x56, 0x5D, 0x01, 0x33, 0xFD, 0xF4, 0xAE,
		0xDE, 0x30, 0x07, 0x9B, 0xE5, 0x83, 0x9B, 0x68,
		0x49, 0xB4, 0x2E, 0x83, 0x1F, 0xC2, 0xB5, 0x7C,
		0xA2, 0x19, 0xD8, 0xE5, 0x7C, 0x2F, 0x83, 0xDA,
		0xF7, 0x6B, 0x90, 0xFE, 0xC4, 0x01, 0x5A, 0x97,
		0x61, 0xA6, 0x3D, 0x40, 0x0B, 0x58, 0xE6, 0x3D
	}, {
		0x4D, 0xD1, 0xB2, 0x0F, 0x28, 0xBD, 0xE4, 0x78,
		0xF6, 0x4A, 0x0F, 0x93, 0x8B, 0x17, 0xD1, 0xA4,
		0x3A, 0xEC, 0xC9, 0x35, 0x93, 0x56, 0x7E, 0xCB,
		0x55, 0x20, 0xA0, 0xFE, 0x6C, 0x89, 0x17, 0x62,
		0x17, 0x62, 0x4B, 0xB1, 0xB4, 0xDE, 0xD1, 0x87,
		0xC9, 0x14, 0x3C, 0x4A, 0x7E, 0xA8, 0xE2, 0x7D,
		0xA0, 0x9F, 0xF6, 0x5C, 0x6A, 0x09, 0x8D, 0xF0,
		0x0F, 0xD3, 0x53, 0x25, 0x95, 0x36, 0x28, 0xCB
	}
};

void des_crypt(unsigned char to[BLOCK_SIZE],
               const unsigned char from[BLOCK_SIZE],
               unsigned char subkey[NUM_ROUNDS][6])
{
	unsigned char block[BLOCK_SIZE];
	unsigned char *left, *right;
	unsigned int round;
	unsigned char temp[BLOCK_SIZE / 2];

	init_perm(block, from);
	left = block;
	right = block + (BLOCK_SIZE / 2);
	for (round = 0; round < NUM_ROUNDS; round++) {
		des_round(left, right, subkey[round]);
	}
	memcpy(temp, left, 4);
	memcpy(left, right, 4);
	memcpy(right, temp, 4);
	final_perm(to, block);

	return;
}

void des_round(unsigned char left[BLOCK_SIZE / 2],
               unsigned char right[BLOCK_SIZE / 2],
               unsigned char subkey[6])
{
	unsigned char left_save[BLOCK_SIZE / 2], temp[BLOCK_SIZE / 2];
	unsigned char expand[6];

	memcpy(left_save, left, BLOCK_SIZE / 2);
	memcpy(left, right, BLOCK_SIZE / 2);
	exp_perm(expand, right);
	xor(expand, subkey, 6);
	s_box_sub(temp, expand);
/*
	bit_permute(right, temp, p_box_perm, 32);
*/
	p_box_perm((unsigned long *) right, (unsigned long *) temp);
	xor(right, left_save, BLOCK_SIZE / 2);

	return;
}

int des_key_init(unsigned char subkey[NUM_ROUNDS][6],
                 unsigned char long_key[LONG_KEY_SIZE], int crypt)
{
	void (*key_shift)(unsigned char[KEY_SIZE], unsigned int num);
	static unsigned int bit_shift[2][NUM_ROUNDS] = {
		{ 1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1 },
		{ 0, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1 }
	};
	unsigned char temp_key[KEY_SIZE];
	unsigned int i;

	crypt = (crypt != 0);	/* Make sure it's 1 or 0 */
	key_shift = (crypt == DES_ENCRYPT) ? key_shift_left : key_shift_right;
	bit_permute(temp_key, long_key, key_perm, 56);
	for (i = 0; i < NUM_ROUNDS; i++) {
		key_shift(temp_key, bit_shift[crypt][i]);
		bit_permute(subkey[i], temp_key, comp_perm, 48);
	}

	return 0;
}

void bit_permute(unsigned char *to, const unsigned char *from,
                 const unsigned char *perm, unsigned int num)
{
	const unsigned char mask_array[8] = {
		0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
	};
	unsigned char to_mask;
	int i;

	to_mask = 0x80;
	*to = 0x00;
	for (i = 0; i < num; i++) {
		if (to_mask == 0x00) {
			to_mask = 0x80;
			*(++to) = 0x00;
		}
		if (from[*perm >> 3] & mask_array[*perm & 0x07])
			*to |= to_mask;
		to_mask >>= 1;
		perm++;
	}

	return;
}

void init_perm(unsigned char to[BLOCK_SIZE],
               const unsigned char from[BLOCK_SIZE])
{
	unsigned char to_mask, from_mask;

	from_mask = 0x40;
	while (from_mask != 0x00) {
		from += BLOCK_SIZE;
		to_mask = 0x80;
		*to = 0x00;
		while (to_mask != 0x00) {
			if (from_mask & *(--from))
				*to |= to_mask;
			to_mask >>= 1;
		}
		to++;
		from_mask >>= 2;
	}
	from_mask = 0x80;
	while (from_mask != 0x00) {
		from += BLOCK_SIZE;
		to_mask = 0x80;
		*to = 0x00;
		while (to_mask != 0x00) {
			if (from_mask & *(--from))
				*to |= to_mask;
			to_mask >>= 1;
		}
		to++;
		from_mask >>= 2;
	}

	return;
}

void final_perm(unsigned char to[BLOCK_SIZE],
                const unsigned char from[BLOCK_SIZE])
{
	unsigned char to_mask, from_mask;

	from_mask = 0x01;
	while (from_mask != 0x00) {
		to_mask = 0x80;
		*to = 0x00;
		while (to_mask != 0x00) {
			if (from_mask & *(from + 4))
				*to |= to_mask;
			to_mask >>= 1;
			if (from_mask & *from)
				*to |= to_mask;
			to_mask >>= 1;
			from++;
		}
		to++;
		from -= 4;
		from_mask <<= 1;
	}

	return;
}

/*
int key_perm(unsigned char to[KEY_SIZE],
             const unsigned char from[LONG_KEY_SIZE])
{
	unsigned char to_mask, from_mask;

	from_mask = 0x80;
	while (from_mask != 0x10) {
		from += LONG_KEY_SIZE;
		to_mask = 0x80;
		*to = 0x00;
		while (to_mask != 0x00) {
			if (from_mask & *(--from))
				*to |= to_mask;
			to_mask >>= 1;
		}
		to++;
		from_mask >>= 1;
	}

	return 0;
}
*/

void exp_perm(unsigned char to[6], const unsigned char from[4])
{
	to[0] = ((from[0] >> 3) & 0x03) | ((from[0] >> 1) & 0x7C) | (from[3] << 7);
	to[1] = ((from[0] << 3) & 0x08) | (from[0] << 5) |
	        ((from[1] >> 3) & 0x10) | (from[1] >> 5);
	to[2] = ((from[1] << 3) & 0xC0) | ((from[1] << 1) & 0x3E) | (from[2] >> 7);
	to[3] = ((from[2] >> 3) & 0x03) | ((from[2] >> 1) & 0x7C) | (from[1] << 7);
	to[4] = ((from[2] << 3) & 0x08) | (from[2] << 5) |
	        ((from[3] >> 3) & 0x10) | (from[3] >> 5);
	to[5] = ((from[3] << 3) & 0xC0) | ((from[3] << 1) & 0x3E) | (from[0] >> 7);

	return;
}

void s_box_sub(unsigned char to[BLOCK_SIZE / 2], unsigned char from[6])
{
	unsigned char temp;

	temp = s_box[0][from[0] >> 2];
	to[0] = temp & 0xF0;
	temp = s_box[0][((from[0] << 4) | (from[1] >> 4)) & 0x3F];
	to[0] |= temp & 0x0F;
	temp = s_box[1][((from[1] << 2) | (from[2] >> 6)) & 0x3F];
	to[1] = temp & 0xF0;
	temp = s_box[1][from[2] & 0x3F];
	to[1] |= temp & 0x0F;
	temp = s_box[2][from[3] >> 2];
	to[2] = temp & 0xF0;
	temp = s_box[2][((from[3] << 4) | (from[4] >> 4)) & 0x3F];
	to[2] |= temp & 0x0F;
	temp = s_box[3][((from[4] << 2) | (from[5] >> 6)) & 0x3F];
	to[3] = temp & 0xF0;
	temp = s_box[3][from[5] & 0x3F];
	to[3] |= temp & 0x0F;

	return;
}

void p_box_perm(unsigned long *to,
                const unsigned long *from)
{
	static unsigned long mask[32] = {
		1 << 15, 1 <<  6, 1 << 19, 1 << 20, 1 << 28, 1 << 11, 1 << 27, 1 << 16,
		1 <<  0, 1 << 14, 1 << 22, 1 << 25, 1 <<  4, 1 << 17, 1 << 30, 1 <<  9,
		1 <<  1, 1 <<  7, 1 << 23, 1 << 13, 1 << 31, 1 << 26, 1 <<  2, 1 <<  8,
		1 << 18, 1 << 12, 1 << 29, 1 <<  5, 1 << 21, 1 << 10, 1 <<  3, 1 << 24
	};
	unsigned long *to_mask;
	unsigned long from_mask;

	*to = 0x00000000UL;
	to_mask = mask;
	from_mask = 0x80000000UL;
	while (from_mask != 0) {
		if (*from & from_mask)
			*to |= *to_mask;
		from_mask >>= 1;
		to_mask++;
	}

	return;
}

void key_shift_left(unsigned char key[KEY_SIZE], unsigned int num)
{
	unsigned char masks[5] = {
		0x00, 0x80, 0xC0, 0xE0, 0xF0
	};
	unsigned char save, mask;
	unsigned int i;

	mask = masks[num];
	key += KEY_SIZE;
	save = 0x00;
	for (i = KEY_SIZE; i > 0; i--) {
		key--;
		save |= *key & mask;
		*key <<= num;
		*key |= save & ~mask;
		save >>= 8 - num;
	}
	mask = masks[4 - num];
	key[6] |= (key[3] & ~mask) >> 4;
	key[3] &= mask | 0x0F;
	key[3] |= save << 4;

	return;
}

void key_shift_right(unsigned char key[KEY_SIZE], unsigned int num)
{
	unsigned char masks[5] = {
		0x00, 0x01, 0x03, 0x07, 0x0F
	};
	unsigned char save, mask;
	unsigned int i;

	mask = masks[num];
	save = 0x00;
	for (i = 0; i < KEY_SIZE; i++) {
		save |= *key & mask;
		*key >>= num;
		*key |= save & ~mask;
		save <<= 8 - num;
		key++;
	}
	mask = masks[4 - num];
	key -= KEY_SIZE;
	key[0] |= (key[3] & ~mask) << 4;
	key[3] &= mask | 0xF0;
	key[3] |= save >> 4;

	return;
}

void xor(void *s, const void *t, size_t num)
{
/*
	unsigned long *ls, *lt;

	ls = s;
	lt = t;
	*ls++ ^= *lt++;
	*ls ^= *lt;

	return;
*/
	unsigned int *is, *it;
	unsigned char *cs, *ct;

	is = (unsigned int *) s;
	it = (unsigned int *) t;
	while (num >= sizeof(unsigned int)) {
		*is++ ^= *it++;
		num -= sizeof(int);
	}
	cs = (unsigned char *) is;
	ct = (unsigned char *) it;
	while (num > 0) {
		*cs++ ^= *ct++;
		num--;
	}

	return;
}

