/*
	des_mode.c
	Copyright (C) 2000 David Fifield

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include "des_mode.h"
#include "des.h"

static void get_iv(unsigned char to[BLOCK_SIZE])
{
	time_t t;

	time(&t);
	p_box_perm((unsigned long *) to, (unsigned long *) &t);

	return;
}

int des_mode_ecb(FILE *out_file, FILE *in_file,
                 unsigned char key[LONG_KEY_SIZE], unsigned int crypt_dir)
{
	unsigned char block[2][BLOCK_SIZE];
	unsigned char *block_prev, *block_cur, *temp;
	unsigned char subkey[NUM_ROUNDS][SUBKEY_SIZE];
	int count;

	des_key_init(subkey, key, crypt_dir);
	memset(key, 0, LONG_KEY_SIZE);

	block_cur = block[0];
	block_prev = block[1];
	count = fread(block_prev, 1, BLOCK_SIZE, in_file);

	if (feof(in_file) != 0) {
		if (crypt_dir == DES_ENCRYPT) {
			/* Add padding */
			memset(block_prev + count, '\0', BLOCK_SIZE - count);
			block_prev[BLOCK_SIZE - 1] = BLOCK_SIZE - count;
			des_crypt(block_prev, block_prev, subkey);
			if (fwrite(block_prev, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
				return 1;
			return 0;
		} else {
			return 2;
		}
	}

	count = fread(block_cur, 1, BLOCK_SIZE, in_file);

	while (feof(in_file) == 0) {
		des_crypt(block_prev, block_prev, subkey);
		if (fwrite(block_prev, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
			return 1;
		/* Swap the buffers */
		temp = block_cur;
		block_cur = block_prev;
		block_prev = temp;
		count = fread(block_cur, 1, BLOCK_SIZE, in_file);
	}

	if (crypt_dir == DES_ENCRYPT) {
		des_crypt(block_prev, block_prev, subkey);
		if (fwrite(block_prev, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
			return 1;
		/* Add padding */
		memset(block_cur + count, '\0', BLOCK_SIZE - count);
		block_cur[BLOCK_SIZE - 1] = BLOCK_SIZE - count;
		des_crypt(block_cur, block_cur, subkey);
		if (fwrite(block_cur, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
			return 1;
	} else {
		/* Remove padding */
		if (count != 0)
			return 2;
		des_crypt(block_prev, block_prev, subkey);
		count = block_prev[BLOCK_SIZE - 1];
		count = BLOCK_SIZE - count;
		if (fwrite(block_prev, 1, count, out_file) != count)
			return 1;
	}

	memset(subkey, 0, NUM_ROUNDS * SUBKEY_SIZE);

	return 0;
}

int des_mode_cbc(FILE *out_file, FILE *in_file,
                 unsigned char key[LONG_KEY_SIZE], unsigned int crypt_dir)
{
	unsigned char block[4][BLOCK_SIZE];
	unsigned char subkey[NUM_ROUNDS][SUBKEY_SIZE];
	int count;

	des_key_init(subkey, key, crypt_dir);
	memset(key, 0, LONG_KEY_SIZE);

	if (crypt_dir == DES_ENCRYPT) {
		unsigned char *block_prev, *block_cur, *temp;
		block_prev = block[0];
		block_cur = block[1];

		get_iv(block_prev);

		for (;;) {
			count = fread(block_cur, 1, BLOCK_SIZE, in_file);
			if (feof(in_file) != 0)
				break;
			xor(block_cur, block_prev, BLOCK_SIZE);
			des_crypt(block_cur, block_cur, subkey);
			if (fwrite(block_prev, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
				return 1;
			/* Swap the buffers */
			temp = block_cur;
			block_cur = block_prev;
			block_prev = temp;
		}

		/* Ciphertext stealing */
		memset(block_cur + count, 0, BLOCK_SIZE - count);
		xor(block_cur, block_prev, BLOCK_SIZE);
		des_crypt(block_cur, block_cur, subkey);
		if (fwrite(block_cur, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
			return 1;
		if (fwrite(block_prev, 1, count, out_file) != count)
			return 1;
	} else {
		fread(block[1], 1, BLOCK_SIZE, in_file);
		if (feof(in_file) != 0)
			return 2;
		count = fread(block[2], 1, BLOCK_SIZE, in_file);
		if (feof(in_file) != 0) {
			memset(block[2] + count, 0, BLOCK_SIZE - count);
			des_crypt(block[0], block[1], subkey);
			xor(block[0], block[2], BLOCK_SIZE);
			if (fwrite(block[0], 1, count, out_file) != count)
				return 1;
			return 0;
		}

		for (;;) {
			des_crypt(block[0], block[2], subkey);
			xor(block[0], block[1], BLOCK_SIZE);
			count = fread(block[3], 1, BLOCK_SIZE, in_file);
			if (feof(in_file) != 0)
				break;
			if (fwrite(block[0], 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
				return 1;
			/* Rotate the buffers */
			memmove(block[0], block[1], 3 * BLOCK_SIZE);
		}

		/* Rotate the buffers */
		memmove(block[0], block[1], 3 * BLOCK_SIZE);
		/* Ciphertext stealing */
		memset(block[2] + count, 0, BLOCK_SIZE - count);
		des_crypt(block[1], block[1], subkey);
		xor(block[1], block[2], BLOCK_SIZE);
		memcpy(block[2] + count, block[1] + count, BLOCK_SIZE - count);
		des_crypt(block[2], block[2], subkey);
		xor(block[2], block[0], BLOCK_SIZE);
		if (fwrite(block[2], 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
			return 1;
		if (fwrite(block[1], 1, count, out_file) != count)
			return 1;
	}

	memset(subkey, 0, NUM_ROUNDS * SUBKEY_SIZE);

	return 0;
}

int des_mode_ofb(FILE *out_file, FILE *in_file,
                 unsigned char key[LONG_KEY_SIZE], unsigned int crypt_dir)
{
	unsigned char subkey[NUM_ROUNDS][SUBKEY_SIZE];
	unsigned char shift_reg[BLOCK_SIZE], block[BLOCK_SIZE];
	unsigned char c;
	int i;

	if (crypt_dir == DES_ENCRYPT) {
		get_iv(shift_reg);
		if (fwrite(shift_reg, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
			return 1;
	} else {
		fread(shift_reg, 1, BLOCK_SIZE, in_file);
		if (feof(in_file) != 0)
			return 2;
	}

	des_key_init(subkey, key, DES_ENCRYPT);
	memset(key, 0, LONG_KEY_SIZE);

	while ((i = getc(in_file)) != EOF) {
		c = (unsigned char) i;
		des_crypt(block, shift_reg, subkey);
		memmove(shift_reg, shift_reg + 1, BLOCK_SIZE - 1);
		c ^= *block;
		if (putc(c, out_file) == EOF)
			return 1;
		shift_reg[BLOCK_SIZE - 1] = *block;
	}

	memset(subkey, 0, NUM_ROUNDS * SUBKEY_SIZE);

	return 0;
}

int des_mode_cfb(FILE *out_file, FILE *in_file,
                 unsigned char key[LONG_KEY_SIZE], unsigned int crypt_dir)
{
	unsigned char subkey[NUM_ROUNDS][SUBKEY_SIZE];
	unsigned char shift_reg[BLOCK_SIZE], block[BLOCK_SIZE];
	unsigned char c;
	int i;

	if (crypt_dir == DES_ENCRYPT) {
		get_iv(shift_reg);
		if (fwrite(shift_reg, 1, BLOCK_SIZE, out_file) != BLOCK_SIZE)
			return 1;
	} else {
		fread(shift_reg, 1, BLOCK_SIZE, in_file);
		if (feof(in_file) != 0)
			return 2;
	}

	des_key_init(subkey, key, DES_ENCRYPT);
	memset(key, 0, LONG_KEY_SIZE);

	while ((i = getc(in_file)) != EOF) {
		c = (unsigned char) i;
		des_crypt(block, shift_reg, subkey);
		memmove(shift_reg, shift_reg + 1, BLOCK_SIZE - 1);
		if (crypt_dir != DES_ENCRYPT)
			block[BLOCK_SIZE - 1] = c;
		c ^= *block;
		if (crypt_dir == DES_ENCRYPT)
			block[BLOCK_SIZE - 1] = c;
		if (putc(c, out_file) == EOF)
			return 1;
		shift_reg[BLOCK_SIZE - 1] = *block;
	}

	memset(subkey, 0, NUM_ROUNDS * SUBKEY_SIZE);

	return 0;
}

